#![no_main]
#![no_std]

use core::ffi::c_void;

use log::{error, info};
use pe_parser::section::section_header;
use uefi::{
    boot::{
        image_handle, install_protocol_interface, load_image, open_protocol_exclusive, start_image,
        LoadImageSource, ScopedProtocol,
    },
    prelude::*,
    proto::{device_path::FfiDevicePath, loaded_image::LoadedImage},
    CString16, Error, Guid,
};

// Not used directly, see INITRD_DEVICE_PATH below.
// const EFI_INITRD_MEDIA_GUID: Guid = Guid::parse_or_panic("5568e427-68fc-4f3d-ac74-ca555231cc68");

/// GUID for the `EFI_LOAD_FILE2_PROTOCOL`.
///
/// See: <https://uefi.org/specs/UEFI/2.10/13_Protocols_Media_Access.html#load-file-2-protocol>
const EFI_LOADFILE2_PROTOCOL: Guid = Guid::parse_or_panic("4006c0c1-fcb3-403e-996d-4a6c8724e06d");

/// GUID for the `EFI_DEVICE_PATH_PROTOCOL`.
///
/// See: <https://uefi.org/specs/UEFI/2.10/10_Protocols_Device_Path_Protocol.html>
const EFI_DEVICE_PATH_PROTOCOL: Guid = Guid::parse_or_panic("09576e91-6d3f-11d2-8e39-00a0c969723b");

/// Unique device path used for the virtual device which contains the initrd.
///
/// Linux will locate a device with this path which implements `EFI_LOAD_FILE2_PROTOCOL`, and use it
/// to load the initrd.
///
/// See: <https://uefi.org/specs/UEFI/2.10/10_Protocols_Device_Path_Protocol.html#efi-device-path-protocol>
/// See: `linux::drivers/firmware/efi/libstub/efi-stub-helper.c`
#[rustfmt::skip]
static mut LINUX_INITRD_DEVICE_PATH: [u8; 24] = [
    // First node in device path.
    0x04, // Type: Media device
    0x03, // Subtype: Vendor-Defined
    0x14, 0x00, // Length: 20 bytes
    // Vendor GUI, EFI_INITRD_MEDIA_GUID, in Microsoft's mixed-endian encoding.
    0x27, 0xe4, 0x68, 0x55,
    0xfc, 0x68,
    0x3d, 0x4f,
    0xac, 0x74,
    0xca, 0x55, 0x52, 0x31, 0xcc, 0x68,
    // Second node in path.
    0x7f, // Type: End of Hardware Device Path
    0xff, // Subtype: End Entire Device Path
    0x04, 0x00, // Length: 4 bytes
];

/// Implementation of the `EFI_LOAD_FILE2_PROTOCOL` protocol
///
/// Used to obtain files from arbitrary devices that are not boot options.
#[repr(C)]
// #[uefi::proto::unsafe_protocol(EFI_LOADFILE2_PROTOCOL)] // XXX: Not needed?
pub struct LoadFile2Protocol {
    pub load_file: LoadFile2,
    initrd: &'static [u8],
}

/// Causes the driver to load a specified file.
pub type LoadFile2 = extern "C" fn(
    this: *const LoadFile2Protocol,
    file_path: *const FfiDevicePath,
    boot_policy: bool,
    buffer_size: *mut usize,
    buffer_ptr: *mut c_void,
) -> Status;

/// Create a new instance from a given PE section header.
///
/// # Safety
///
/// The caller is responsible for ensuring that the section address and size are consistent
/// with the PE image at `base`. The image data must have a static lifetime.
///
/// # Lifetime
///
/// Memory for these sections must have a 'static lifetime.
///
/// # Panics
///
/// If any of the sizes in the header don't fit into [`usize`].
fn section_as_slice(base: *const u8, section: section_header) -> &'static [u8] {
    let offset = section
        .virtual_address
        .try_into()
        .expect("virtual address must fit into usize");
    let size = section
        .virtual_size
        .try_into()
        .expect("virtual size must fit into usize");

    // SAFETY: Caller is responsible for passing a PE section header which matches `base`.
    unsafe { core::slice::from_raw_parts(base.add(offset), size) }
}

/// Data wrapper that works like a named tuple.
struct Components<'a> {
    kernel: &'a [u8],
    cmdline: &'a [u8],
    initrd: &'a [u8],
}

impl Components<'_> {
    fn from_image(image: &ScopedProtocol<LoadedImage>) -> uefi::Result<Self, &'static str> {
        let mut kernel = Option::<&[u8]>::None;
        let mut cmdline = Option::<&[u8]>::None;
        let mut initrd = Option::<&[u8]>::None;

        let (base, size) = image.info();
        info!("Current image is at {:?} and {:?} bytes", base, size);
        let base = base.cast::<u8>();
        let size = size.try_into().expect("image size must fit into usize");
        let ptr = unsafe { core::slice::from_raw_parts(base, size) };

        let parsed = pe_parser::pe::parse_portable_executable(ptr).unwrap();
        for section in parsed.section_table {
            match &section.name {
                b".linux\0\0" => kernel = Some(section_as_slice(base, section)),
                b".cmdline" => cmdline = Some(section_as_slice(base, section)),
                b".initrd\0" => initrd = Some(section_as_slice(base, section)),
                _ => {} // TODO: .ucode? .dtb?
            };
        }

        let Some(cmdline) = cmdline else {
            return Err(Error::new(Status::NOT_FOUND, "cmdline missing from stub"));
        };
        let Some(kernel) = kernel else {
            return Err(Error::new(Status::NOT_FOUND, "kernel missing from stub"));
        };
        let Some(initrd) = initrd else {
            return Err(Error::new(Status::NOT_FOUND, "initrd missing from stub"));
        };

        Ok(Components {
            kernel,
            cmdline,
            initrd,
        })
    }
}

#[entry]
fn main() -> Status {
    uefi::helpers::init().expect("Initialising UEFI services once must succeed.");
    info!("Hello world!");

    let image = image_handle();
    let loaded_image = match open_protocol_exclusive::<LoadedImage>(image) {
        Ok(i) => i,
        Err(err) => {
            let status = err.status();
            match status {
                Status::UNSUPPORTED => {
                    error!("boot services does not support loaded_image protocol");
                }
                Status::ACCESS_DENIED => {
                    error!("boot services denied exclusive access to loaded_image protocol");
                }
                _ => {
                    error!("Unexpected error accessing loaded_image for current image: {err}");
                }
            };
            return status;
        }
    };

    let Components {
        kernel,
        cmdline,
        initrd,
    } = Components::from_image(&loaded_image).unwrap();

    install_initrd_loader(&mut LoadFile2Protocol { load_file, initrd });

    // FIXME: don't copy the entire kernel; return the existing in-memory copy
    //  TODO: maybe use a pe-loader?
    //  TODO: maybe create a new handle and register it as a LoadedImage? unsure if possible.
    //  TODO: consider using the initrdmem= param; it's a pointer to the existing in-memory
    //        initrd without having to make a copy of it into another buffer.
    //        See: https://www.kernel.org/doc/html/latest/admin-guide/kernel-parameters.html

    let kernel_image = load_image(
        image,
        LoadImageSource::FromBuffer {
            buffer: kernel,
            file_path: Some(loaded_image.file_path().unwrap()),
        },
    )
    .expect("Kernel loads into new buffer.");
    info!("kernel image loaded!");

    drop(loaded_image);
    let mut kernel_loaded_image = open_protocol_exclusive::<LoadedImage>(kernel_image).unwrap();

    // The kernel expects its cmdline in UTF-16.
    // TODO: encode the cmdline when building the UEFI bundle, not at boot time.

    let cmdline: CString16 = core::str::from_utf8(cmdline)
        .expect("cmdline must be valid utf-8")
        .try_into()
        .expect("cmdline must be valid unicode with no null characters");
    let cmdline = cmdline.as_bytes();
    let con_size = cmdline
        .len()
        .try_into()
        .expect("cmdline length must fit in u32");
    let con_ptr = cmdline.as_ptr();

    // SAFETY: con_ptr and con_size define an owned portion of 'static memory.
    unsafe { kernel_loaded_image.set_load_options(con_ptr, con_size) };

    start_image(kernel_image).expect("Kernel image should have started.");

    Status::SUCCESS
}

fn install_initrd_loader(lf: &mut LoadFile2Protocol) {
    let lf = core::ptr::from_mut::<LoadFile2Protocol>(lf).cast::<c_void>();
    // SAFETY: GUID is valid for this protocol.
    let handle = unsafe { install_protocol_interface(None, &EFI_LOADFILE2_PROTOCOL, lf) }
        .expect("Installing LoadFile2Protocol must succeed.");

    #[allow(static_mut_refs)]
    // SAFETY: this is the only reference taken to this constant.
    let dp = unsafe { &mut LINUX_INITRD_DEVICE_PATH };
    let dp = core::ptr::from_mut(dp).cast::<c_void>();
    // SAFETY: GUID is valid for this protocol.
    unsafe { install_protocol_interface(Some(handle), &EFI_DEVICE_PATH_PROTOCOL, dp) }
        .expect("Installing DevicePathProtocol must succeed.");
}

/// Load a file. Only the `initrd` shall be loaded.
///
/// This is called by the kernel to load the initramfs.
///
/// # Safety
///
/// The caller is responsible for ensuring that all pointer are valid.
extern "C" fn load_file(
    this: *const LoadFile2Protocol,
    _file_path: *const FfiDevicePath,
    boot_policy: bool,
    buffer_size: *mut usize,
    buffer_ptr: *mut c_void,
) -> Status {
    assert!(!this.is_null(), "Null pointer for EFI_LOAD_FILE2.This");
    // SAFETY: Caller is responsible for ensuring that the pointer is valid.
    let this = unsafe { &*this };

    // SAFETY: caller is responsible for this pointer being valid.
    let buffer_size = unsafe { buffer_size.as_mut() };
    let Some(buffer_size) = buffer_size else {
        return Status::INVALID_PARAMETER;
    };
    if boot_policy {
        return Status::UNSUPPORTED;
    }

    let file_size = this.initrd.len();
    if *buffer_size < file_size || buffer_ptr.is_null() {
        *buffer_size = file_size;
        return Status::BUFFER_TOO_SMALL;
    }

    // SAFETY: caller is responsible for passing a valid pointer and size.
    let buffer = unsafe { core::slice::from_raw_parts_mut(buffer_ptr.cast::<u8>(), file_size) };

    buffer.copy_from_slice(this.initrd);
    Status::SUCCESS
}
